;***********************************************************************************;
;***********************************************************************************;
;
; A re-implementation of the VIC-20 RS-232 KERNAL routines.
; By Simon Rowe <srowe@mose.org.uk>.
;
; The implementation of X-line handshaking in the KERNAL is intended for use with
; legacy half-duplex modems. This wedge implements the more modern, symmetric, RTS/
; CTS handshaking that allows reliable DTE to DTE communication.
;
;***********************************************************************************;
;***********************************************************************************;


; KERNAL zero page

STATUS	= $90			; I/O status byte
XSAV	= $97			; register save
LDTND	= $98			; open file count
DFLTN	= $99			; input device number
DFLTO	= $9A			; output device number
TIME	= $A0			; jiffy clock high byte
INBIT	= $A7			; tape write leader count/block count/RS-232 input bit
BITCI	= $A8			; tape error flags/tape long word marker/RS-232 input bit count
RINONE	= $A9			; tape dipole count/tape medium word marker/RS-232 start bit flag
RIDATA	= $AA			; tape input status/tape sync status/RS-232 byte assembly
BITTS	= $B4			; transmitter bit count out
NXTBIT	= $B5			; transmitter next bit to be sent
RODATA	= $B6			; transmitter byte buffer/disassembly location
FNLEN	= $B7			; file name length
LA	= $B8			; logical file
SA	= $B9			; secondary address
FA	= $BA			; current device number
FNADR	= $BB			; file name pointer
RIBUF	= $F7			; RS-232 Rx pointer
ROBUF	= $F9			; RS-232 Tx pointer

LAT	= $0259			; .. to $0262 logical file table
FAT	= $0263			; .. to $026C device number table
SAT	= $026D			; .. to $0276 secondary address table
SHFLAG	= $028D			; keyboard shift/control flag
M51CTR	= $0293			; pseudo 6551 control register. the first character of
					; the OPEN RS-232 filename will be stored here
					; bit	function
					; ---	--------
					; 7654	unused
					; 3210	baud rate
					; ----	---------
					; 0000	user rate *
					; 0001	   50
					; 0010	   75
					; 0011	  110
					; 0100	  134.5
					; 0101	  150
					; 0110	  300
					; 0111	  600
					; 1000	 1200
					; 1001	 1800
					; 1010	 2400
					; 1011	 3600
					; 1100	 4800 *
					; 1101	 7200 *
					; 1110	 9600 *
					; 1111	19200 * * = not implemented
M51CDR	= $0294			; pseudo 6551 command register. the second character of
					; the OPEN RS-232 filename will be stored here
					; bit	function
					; ---	--------
					; 765	unused
					;  4	unused
					;  3	unused
					;  2	unused
					;  1	unused
					;  0	handshake - X line/3 line
RSSTAT	= $0297			; RS-232 status register
					; bit	function
					; ---	--------
					;  7	break
					;  6	no DSR detected
					;  5	unused
					;  4	no CTS detected
					;  3	unused
					;  2	Rx buffer overrun
					;  1	framing error
					;  0	parity error
BAUDOF	= $0299			; time of one bit cell
RIDBE	= $029B			; index to Rx buffer end
RIDBS	= $029C			; index to Rx buffer start
RODBS	= $029D			; index to Tx buffer start
RODBE	= $029E			; index to Tx buffer end

VECTORS	= $031A			; KERNAL vector table
NMINV	= $0318			; NMI vector

VIA1PB		= $9110		; VIA 1 DRB
					; bit	function
					; ---	--------
					;  7	DSR in
					;  6	CTS in
					;  5	unused
					;  4	DCD in
					;  3	RI in
					;  2	DTR out
					;  1	RTS out
					;  0	data in
VIA1PA1		= $9111		; VIA 1 DRA
					; bit	function
					; ---	--------
					;  7	serial ATN out
					;  6	cassette switch
					;  5	joystick fire, light pen
					;  4	joystick left, paddle X fire
					;  3	joystick down
					;  2	joystick up
					;  1	serial DATA in
					;  0	serial CLK in
VIA1DDRB	= $9112		; VIA 1 DDRB
VIA1T1CL	= $9114		; VIA 1 T1C_l
VIA1T1CH	= $9115		; VIA 1 T1C_h
VIA1T2CL	= $9118		; VIA 1 T2C_l
VIA1T2CH	= $9119		; VIA 1 T2C_h
VIA1PCR		= $911C		; VIA 1 PCR
					; bit	function
					; ---	--------
					; 765	CB2 Tx RS-232 data
					;  4	CB1 Rx RS-232 data
					; 321	CA2 cassette motor control
					;  0	CA1 [RESTORE] key

; The status bit is a not normal flag. It goes high if both an interrupt flag in the IFR
; and the corresponding enable bit in the IER are set. It can be cleared only by clearing
; all the active flags in the IFR or disabling all active interrupts in the IER.

VIA1IFR		= $911D		; VIA 1 IFR
					; bit	function		cleared by
					; ---	--------		----------
					;  7	interrupt status	clearing all enabled interrupts
					;  6	T1 interrupt		read T1C_l, write T1C_h
					;  5	T2 interrupt		read T2C_l, write T2C_h
					;  4	CB1 transition		read or write port B
					;  3	CB2 transition		read or write port B
					;  2	8 shifts done		read or write the shift register
					;  1	CA1 transition		read or write port A
					;  0	CA2 transition		read or write port A

; If enable/disable bit is a zero during a write to this register, each 1 in bits 0-6
; clears the corresponding bit in the IER. If this bit is a one during a write to this
; register, each 1 in bits 0-6 will set the corresponding IER bit.

VIA1IER		= $911E		; VIA 1 IER
					; bit	function
					; ---	--------
					;  7	enable/disable
					;  6	T1 interrupt, Tx timer
					;  5	T2 interrupt, Rx timer
					;  4	CB1 transition, Rx start bit
					;  3	CB2 transition
					;  2	8 shifts done
					;  1	CA1 transition, RESTORE
					;  0	CA2 transition

	
	; BASIC entrypoints
COLDST	= $C000			; BASIC cold start entry point
WARMST	= $C002			; BASIC warm start entry point

	; KERNAL entrypoints
INITSK	= $E518			; initialize hardware
FGETIN	= $F1F5			; get character from keyboard queue
FCHRIN	= $F20E			; input character from channel
FCHROUT	= $F27A			; output a character to channel
FOPEN	= $F40A			; open a logical file
FNDFLNO	= $F3CF			; find file
SETFLCH = $F3DF			; set file details from table,.X
FCLRCHN = $F3F3			; close input and output channels
FSTOP	= $F770			; scan stop key
FE_2MNYF = $F77E		; display too many files
FE_ALOPN = $F781		; display file already open
FE_NTOPN = $F784		; display file not open
FE_NTINP = $F78D		; display not input file
FRESTOR	= $FD52			; restore default I/O vectors
INITMEM	= $FD8D			; initialize and test RAM
INITVIA	= $FDF9			; initialize I/O registers
BAUDTBL	= $FF5C			; baud rate tables
MEMTOP	= $FF99			; read/set the top of memory
STOP	= $FFE1			; scan the stop key


FLWCTL	= $AB			; Rx flow control, $00 = disabled, $80 = enabled
UPDSTAT	= $B0			; update RSSTAT flag, $80 = update needed
LNSTAT	= $B1			; CTS and DSR line status
ENABL	= $BD			; NMI interrupts to enable on return
RXBAUD	= $BE			; Rx baud count
RSSTAT2	= $029F			; additional status register, $80 = Tx buffer full


WTRMARK	= $10


*	=	$B000

	.WORD	RSCOLD
	.WORD	0			; unused
	.BYT	"B0",'C'+$80,'B'+$80,'M'+$80


#ifdef TEST_HARNESS
	JMP	TESTH
#endif


;***********************************************************************************;
;
; wedge cold start entry point

RSCOLD
	JSR	INITMEM			; initialize and test RAM
	JSR	FRESTOR			; restore default I/O vectors
	JSR	RSVECT			; set our KERNAL vectors
	JSR	INITVIA			; initialize I/O registers
	JSR	INITSK			; initialize hardware
	CLI				; enable interrupts
	JMP	(COLDST)		; execute BASIC


; replacement KERNAL vectors

VTABLE	.WORD ROPEN
	.WORD RCLOSE
	.WORD RCHKIN
	.WORD RCHKOUT
	.WORD FCLRCHN
	.WORD RCHRIN
	.WORD RCHROUT
	.WORD FSTOP
	.WORD RGETIN


;***********************************************************************************;
;
; set KERNAL and NMI vectors

RSVECT
	LDY	#$11			; set byte count
KLOOP	LDA	VTABLE,Y		; get byte from table
	STA	VECTORS,Y		; set vector
	DEY				; decrement index
	BPL	KLOOP			; loop if more to do
	LDA	#<RFNMI			; get NMI routine low byte
	STA	NMINV			; set NMI vector low byte
	LDA	#>RFNMI			; set NMI routine high byte
	STA	NMINV+1			; set NMI vector high byte
	RTS


;***********************************************************************************;
;
; open a logical file

ROPEN
.(
	LDA	FA			; get device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	FOPEN			; jump to normal KERNAL routine

OVER	LDX	LA			; get logical file
	BNE	ALOP			; branch if there is a file

	JMP	FE_NTINP		; else do not input file error and return

ALOP
	JSR	FNDFLNO			; find file
	BNE	CHKMX			; branch if file not found

	JMP	FE_ALOPN		; else do file already open error and return

CHKMX
	LDX	LDTND			; get open file count
	CPX	#$0A			; compare with max
	BCC	DOOP			; branch if less

	JMP	FE_2MNYF		; else do too many files error and return

DOOP	INC	LDTND			; increment open file count
	LDA	LA			; get logical file
	STA	LAT,X			; save to logical file table
	LDA 	SA			; get secondary address
	ORA	#$60			; OR with the OPEN CHANNEL command
	STA	SA			; set secondary address
	STA	SAT,X			; save to secondary address table
	LDA	FA			; get device number
	STA	FAT,X			; save to device number table

	JMP	OPENRS			; open RS-232 device
.)


;***********************************************************************************;
;
; close a specified logical file

RCLOSE
.(
	JSR	$F3D4			; find file .A
	BEQ	LABEL1			; if the file is found go close it

	CLC				; else the file was closed so just flag ok
	RTS

; found the file so close it

LABEL1
	JSR	SETFLCH			; set file details from table,.X
	LDA	FA			; get device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	$F351			; jump to normal KERNAL routine

OVER	JMP	CLOSERS			; close RS-232 device
.)


;***********************************************************************************;
;
; open a channel for input

RCHKIN
.(
	JSR	FNDFLNO			; find file
	BEQ	LABEL1			; branch if file opened

	JMP	FE_NTOPN		; do file not open error and return

LABEL1
	JSR	SETFLCH			; set file details from table,.X
	LDA	FA			; get device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	$F2D2			; jump to normal KERNAL routine

OVER	JMP	$F2EC			; set input device number
.)


;***********************************************************************************;
;
; open a channel for output

RCHKOUT
.(
	JSR	FNDFLNO			; find file
	BEQ	LABEL1			; branch if file opened

	JMP	FE_NTOPN		; do file not open error and return

LABEL1
	JSR	SETFLCH			; set file details from table,.X
	LDA	FA			; get device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	$F314			; jump to normal KERNAL routine

OVER	JMP	$F32E			; set output device number
.)


;***********************************************************************************;
;
; input character from channel

RCHRIN
.(
	LDA	DFLTN			; get input device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	FCHRIN			; jump to normal KERNAL routine

OVER	JSR	RSNXTIN			; receive byte
	BCC	OVER			; loop until byte received

	CLC				; flag ok
	RTS
.)


;***********************************************************************************;
;
; output a character to channel

RCHROUT
.(
	STY	XSAV			; save .Y
	LDY	DFLTO			; get output device number
	CPY	#$02			; compare with RS-232 device
	BEQ	OVER
	LDY	XSAV			; restore .Y
	JMP	FCHROUT			; jump to normal KERNAL routine

OVER
	JSR	RSOUTSAV		; send byte
	LDY	XSAV			; restore .Y
	CLC				; flag ok
	RTS
.)


;***********************************************************************************;
;
; get a character from the input device

RGETIN
.(
	LDA	DFLTN			; get input device number
	CMP	#$02			; compare with RS-232 device
	BEQ	OVER
	JMP	FGETIN			; jump to normal KERNAL routine

OVER	JMP	RSNXTIN			; receive byte
.)


;***********************************************************************************;
;
; open RS-232 device

OPENRS
	LDA	#$06			; IIII IOOI, DTR and RTS only as outputs
	STA	VIA1DDRB		; set VIA 1 DDRB
	STA	VIA1PB			; set VIA 1 DRB, DTR and RTS high
	LDA	#$EE			; CB2 high, CB1 -ve edge, CA2 high, CA1 -ve edge
	STA	VIA1PCR			; set VIA 1 PCR
	LDA	#$20			; RS-232 Tx high
	STA	NXTBIT			; save RS-232 next bit to send
	LDY	#$00			; clear index
	STY	RSSTAT			; clear RS-232 status byte
	STY	RSSTAT2			; clear RS-232 additional status byte
	STY	UPDSTAT			; clear update RSSTAT flag
LAB_F4D9
	CPY	FNLEN			; compare with file name length
	BEQ	LAB_F4E7		; exit loop if done

	LDA	(FNADR),Y		; get file name byte
	STA	M51CTR,Y		; copy to 6551 register set
	INY				; increment index
	CPY	#$04			; compare with $04
	BNE	LAB_F4D9		; loop if not to 4 yet

LAB_F4E7
	LDA	M51CTR			; get pseudo 6551 control register
	AND	#$0F			; mask 0000 xxxx, baud rate
	ASL				; * 2
	TAX				; copy to index
	LDA	BAUDTBL-2,X		; get timer constant low byte
	STA	RXBAUD			; set Rx baud count low byte
	ASL				; * 2
	TAY				; copy to .Y
	LDA	BAUDTBL-1,X		; get timer constant high byte
	STA	RXBAUD+1		; set Rx baud high byte
	ROL				; * 2
	PHA				; save it
	TYA				; get timer constant low byte back
	ADC	#$C8			; + 200, carry cleared by previous ROL
	STA	BAUDOF			; save bit cell time low byte
	PLA				; restore high byte
	ADC	#$00			; add carry
	STA	BAUDOF+1		; save bit cell time high byte
	LDA	M51CDR			; get pseudo 6551 command register
	LSR				; shift b0 into Cb
	LDA	#%11000000		; CTS and DSR set
	BCC	SETLNST			; branch if 3 line interface

	LDA	VIA1PB			; get VIA 1 DRB
SETLNST
	STA	LNSTAT			; set line status

LAB_F51B
	LDA	RIDBE			; get index to Rx buffer end
	STA	RIDBS			; set index to Rx buffer start, clear Rx buffer
	LDA	RODBE			; get index to Tx buffer end
	STA	RODBS			; set index to Tx buffer start, clear Tx buffer
	SEC
	JSR	MEMTOP			; read the top of memory
	LDA	RIBUF+1			; get RS-232 Rx buffer pointer high byte
	BNE	LAB_F533		; branch if buffer already set

	DEY				; decrement top of memory high byte, 256 byte buffer
	STY	RIBUF+1			; set RS-232 Rx buffer pointer high byte
	STX	RIBUF			; set RS-232 Rx buffer pointer low byte
LAB_F533
	LDA	ROBUF+1			; get RS-232 Tx buffer pointer high byte
	BNE	SETINT			; branch if buffer already set

	DEY				; decrement Rx buffer pointer high byte, 256 byte buffer
	STY	ROBUF+1			; set RS-232 Tx buffer pointer high byte
	STX	ROBUF			; set RS-232 Tx buffer pointer low byte
SETINT
	LDA	VIA1IER			; get VIA 1 IER, enabled interrupts
	ORA	#%10010000		; enable interrupts, CB1 interrupt
	STA	RINONE			; set RS-232 start bit check flag, no start bit received
	STA	FLWCTL			; set Rx flow control
	STA	VIA1IER			; set VIA 1 IER, enabled interrupts
LAB_F53C
	SEC				; non-standard exit, Cb set
	LDA	#$F0			; non-standard exit, $F0 error code
	JMP	$FE7B			; set the top of memory and return


;***********************************************************************************;
;
; close RS-232 device

CLOSERS
	LDA	#$7D			; disable T1, T2, CB1, CB2, SR and CA2
	STA	VIA1IER			; set VIA 1 IER
	LDA	#$00			; set DTR and RTS low
	STA	VIA1PB			; set VIA 1 DRB
	LDA	#$EE			; CB2 high, RS-232 Tx
					; CB1 -ve edge, RS-232 Rx
					; CA2 high, tape motor off
					; CA1 -ve edge, [RESTORE] key
	STA	VIA1PCR			; set VIA 1 PCR
	SEC
	JSR	MEMTOP			; read the top of memory
	LDA	RIBUF+1			; get RS-232 Rx buffer pointer high byte
	BEQ	LAB_F37F		; branch if no RS-232 input buffer

	INY				; else reclaim RS-232 input buffer memory
LAB_F37F
	LDA	ROBUF+1			; get RS-232 Tx buffer pointer high byte
	BEQ	LAB_F384		; branch if no RS-232 output buffer

	INY				; else reclaim RS-232 output buffer memory
LAB_F384
	LDA	#$00			; clear .A
	STA	RIBUF+1			; clear RS-232 Rx buffer pointer high byte
	STA	ROBUF+1			; clear RS-232 Tx buffer pointer high byte
	JMP	LAB_F53C		; go set top of memory and exit


;***********************************************************************************;
;
; Set status register from VIA lines   XXXX unused XXXX

SETSTAT
.(
	AND	#%11000000		; mask off DSR and CTS bits
	TAY				; copy to .Y
	LDA	RSSTAT			; get RS-232 status register
	AND 	#%10101111		; clear no DSR detected and no CTS detected
	STA 	RSSTAT			; set RS-232 status register
	TYA				; restore VIA lines
	BMI 	TSTCTS			; branch to CTS test if DSR present
	LDA 	#%01000000		; no DSR detected
	ORA 	RSSTAT			; OR with RS-232 status register
	STA 	RSSTAT			; set RS-232 status register
TSTCTS	TYA				; restore VIA lines
	ASL				; shift CTS bit in Sb
	BMI 	OUT			; exit if CTS present
	LDA 	#%00010000		; no CTS detected
	ORA 	RSSTAT			; OR with RS-232 status register
	STA 	RSSTAT			; set RS-232 status register
OUT	RTS
.)


;***********************************************************************************;
;
; re-enable interrupts and exit

NMIDONE
	LDA	ENABL			; get enabled interrupts
	ORA	#$80			; enable interrupts
	STA	VIA1IER			; set VIA 1 IER
	PLA				; pull .Y
	TAY				; restore .Y
	PLA				; pull .X
	TAX				; restore .X
	PLA				; restore .A
	RTI


;***********************************************************************************;
;
; NMI handler

RFNMI
	PHA				; save .A
	TXA				; copy .X
	PHA				; save .X
	TYA				; copy .Y
	PHA				; save .Y
	LDA	VIA1IER			; get VIA 1 IER, enabled interrupts
#ifdef ROCKWELL_VIA
	AND	#$7F			; Rockwell VIAs return Cb set
#endif
	STA	ENABL			; save for re-enabling on exit
	LDA	#$7F			; disable all interrupts, clears b7 of IFR and raises NMI
	STA	VIA1IER			; set VIA 1 IER

PROCNMI
	LDA	VIA1IFR			; get VIA 1 IFR, pending interrupts
	AND	ENABL			; mask with enabled interrupts
	BEQ	NMIDONE			; no more pending interrupts

	TAX				; copy to .X

	BIT	UPDSTAT			; test update RSSTAT flag
	BPL	TESTRXTM		; branch if no update needed
	AND	#$10			; mask CB1 interrupt
	BNE	RSTINT			; branch if CB1 interrupt pending
	LDA	ENABL			; get enabled interrupts
	AND	#$10			; mask CB1 interrupt
	BEQ	RSTINT			; branch if CB1 interrupt not enabled

	; byte to Tx, X line handshaking is enabled but we are awaiting start bit

	LDA	VIA1PB			; read VIA 1 DRB, may clear CB1 interrupt
	STA	LNSTAT			; set line status from VIA lines
	AND	#$01			; mask data in bit
	BEQ	RXSTART			; if Rx is now low, start bit arrived while reading status

RSTINT	TXA				; get pending interrupts back
TESTRXTM
	AND	#$20			; mask timer 2 interrupt
	BEQ	TESTRXST		; branch if not timer 2 interrupt

					; was VIA 1 T2 interrupt
	LDA	VIA1PB			; get VIA 1 DRB
	AND	#$01			; mask RS-232 data in
	STA	INBIT			; save RS-232 input bit
	LDA	VIA1T2CL		; get VIA 1 T2C_l, clear interrupt
	SBC	#$16			; adjust by 22 cycles to cover time taken by the
					; six instructions needed to write to the latch
	ADC	BAUDOF			; add baud rate bit time low byte
	STA	VIA1T2CL		; set VIA 1 T2C_l
	LDA	VIA1T2CH		; get VIA 1 T2C_h
	ADC	BAUDOF+1		; add baud rate bit time high byte
	STA	VIA1T2CH		; set VIA 1 T2C_h
	JSR	RSINBIT			; RS-232 Rx

TESTRXST
	TXA				; get pending interrupts back
	AND	#$10			; mask CB1 interrupt
	BEQ	TESTTXST		; branch if not CB1 interrupt

RXSTART					; was CB1 interrupt
	LDA	RXBAUD			; get Rx baud count low byte
	STA	VIA1T2CL		; set VIA 1 T2C_l
	LDA	RXBAUD+1		; get Rx baud count high byte
	STA	VIA1T2CH		; set VIA 1 T2C_h
	LDA	VIA1PB			; read VIA 1 DRB, clear interrupt flag
	STA	LNSTAT			; set line status from VIA lines
	LDA	ENABL			; get enabled interrupts
	ORA	#$20			; enable T2 interrupt
	AND	#$EF			; disable CB1 interrupt
	STA	ENABL			; set enabled interrupts
	LDA	#$09			; set bit count to 9, 8 data + 1 stop bit
	STA	BITCI			; save RS-232 input bit count

TESTTXST
	TXA				; get pending interrupts back
	AND	#$40			; mask timer 1 interrupt
	BEQ	RESTORE			; branch if not timer 1 interrupt

					; was VIA 1 T1 interrupt
	LDA	#$CE			; CB2 low, CB1 -ve edge, CA2 high, CA1 -ve edge
	ORA	NXTBIT			; OR RS-232 next bit to send, sets CB2 high if set
	STA	VIA1PCR			; set VIA 1 PCR
	LDA	VIA1T1CL		; get VIA 1 T1C_l, clear interrupt
	JSR	RSNXTBIT		; RS-232 Tx NMI routine

RESTORE
	TXA				; get pending interrupts back
	AND	#$02			; mask CA1 interrupt
	BNE	STOPREST		; branch CA1 interrupt
	JMP	PROCNMI			; otherwise repeat interrupt processing

STOPREST				; was [RESTORE] key ..
	BIT	VIA1PA1			; test VIA 1 DRA, clear CA1 interrupt
	JSR	STOP			; scan stop key
	BEQ	DOSTOP			; if [STOP] perform reset
	JMP	PROCNMI			; if not [STOP] repeat interrupt processing

DOSTOP
	JSR	FRESTOR			; restore default I/O vectors
	JSR	RSVECT			; restore our vectors
	JSR	INITVIA			; initialise I/O registers
	JSR	INITSK			; initialise hardware
	JMP	(WARMST)		; do BASIC break entry


;***********************************************************************************;
;
; RS-232 Rx NMI

RSINBIT
	LDA	RINONE			; get RS-232 start bit check flag
	BNE	RSSTRBIT		; branch if no start bit received

	DEC	BITCI			; decrement RS-232 input bit count
	BEQ	RSINBYTE		; if the byte is complete go add it to the buffer

	BMI	LAB_F04D		; determine if all stop bits have been received

	LSR	INBIT			; shift RS-232 input bit into Cb
	ROR	RIDATA			; shift Cb into RS-232 byte assembly
	RTS


;***********************************************************************************;
;
; no RS-232 start bit received

RSSTRBIT
	LDA	INBIT			; get RS-232 input bit
	BNE	RSPREPIN		; branch if bit set
	STA	RINONE			; set RS-232 start bit check flag, start bit received
	RTS


;***********************************************************************************;
;
; put byte into receive buffer

RSINBYTE
	SEC
	LDA	RIDBE			; get index to Rx buffer end
	SBC	RIDBS			; subtract index to Rx buffer start
	EOR	#$FF			; complement result
	BEQ	RSOVERUN		; if buffer full go do Rx overrun error
	TAY				; copy space remaining to .Y
	LDA	M51CDR			; get pseudo 6551 command register
	LSR				; shift b0 into Cb
	BCC	RXOK			; branch if 3 line interface

	CPY	#WTRMARK		; compare with low watermark
	BCS	RXOK			; branch if space is above watermark

	; space has fallen below the watermark, drop RTS

	LDA	VIA1PB			; get VIA 1 DRB
	AND	#%11111101		; clear RTS out
	STA	VIA1PB			; set VIA 1 DRB
	LSR	FLWCTL			; clear Rx flow control

RXOK
	LDY	RIDBE			; get index to Rx buffer end
	LDA	RIDATA			; get assembled byte
	STA	(RIBUF),Y		; save received byte to RS-232 Rx buffer
	INC	RIDBE			; increment index to Rx buffer end

LAB_F04D
	LDA	INBIT			; get RS-232 input bit
	BEQ	LAB_F0B3		; bit clear, either BREAK or framing error


;***********************************************************************************;
;
; prepare to receive next input byte

RSPREPIN
	LDA	ENABL			; get enabled interrupts
	ORA	#$10			; enable CB1 interrupt, Rx data bit transition
	AND	#$DF			; disable T2 interrupt
	STA	ENABL			; set enabled interrupts
	LDA	#$90			;
	STA	RINONE			; set RS-232 start bit check flag, no start bit received
	RTS

RSOVERUN
	LDA	#$04			; set Rx overrun error
	.byte	$2C			; makes next line BIT $80A9

RSBREAK
	LDA	#$80			; Rx break error
	.byte	$2C			; makes next line BIT $02A9

RSFRAMER
	LDA	#$02			; Rx frame error
	ORA	RSSTAT			; OR with RS-232 status byte
	STA	RSSTAT			; save RS-232 status byte
	JMP	RSPREPIN		; prepare to receive next input byte

LAB_F0B3
	LDA	RIDATA			; get assembled byte
	BNE	RSFRAMER		; if not break do frame error

	BEQ	RSBREAK			; else do break error, branch always


;***********************************************************************************;
;
; RS-232 Tx NMI routine

RSNXTBIT
	DEC	BITTS			; decrement RS-232 bit count
	BEQ	RSSTOP			; if zero go do stop bit

	BMI	RSNXTBYT		; if -ve go setup next RS-232 Tx byte and return

					; else bit count is non zero and +ve
	LSR	UPDSTAT			; clear update RSSTAT flag
	LSR	RODATA			; shift RS-232 output byte buffer
	LDA	#$00			; set $00 for bit = 0
	BCC	LAB_EFB0		; branch if bit was 0

	ROR				; shift Cb into ..
	ROR
	ROR				; .. CB2 control bit

LAB_EFB0
	STA	NXTBIT			; save RS-232 next bit to send
	RTS

RSSTOP
	LDA	#$20			; stop bit = 1
	STA	NXTBIT			; save RS-232 next bit to send
SETUPDF
	LDA	M51CDR			; get pseudo 6551 command register
	LSR				; shift handshake bit into Cb
	ROR	UPDSTAT			; shift Cb into update RSSTAT flag
	RTS

; setup next RS-232 Tx byte

RSNXTBYT
	LDY	RODBS			; get index to Tx buffer start
	CPY	RODBE			; compare with index to Tx buffer end
	BEQ	LAB_F021		; if all done go disable T1 interrupt and return

	BIT	LNSTAT			; test line status
	BVC	RETRYTX			; if no CTS clear RS-232 bit count and exit

	LDA	#$00			; clear .A
	STA	UPDSTAT			; clear update RSSTAT flag
	STA	NXTBIT			; clear RS-232 next bit to send, start bit = 0
	LDA	#$09			; set bit count to 9, 8 data + 1 stop bit
	STA	BITTS			; set RS-232 bit count
	LDA	(ROBUF),Y		; else get byte from buffer
	STA	RODATA			; save to RS-232 output byte buffer
	INC	RODBS			; increment index to Tx buffer start
	LDA	RSSTAT2			; get RS-232 additional status byte
	AND	#%01111111		; mask Tx buffer full bit
	STA	RSSTAT2			; set RS-232 additional status byte
	RTS

; retry on next timer interrupt

RETRYTX
	LDA	M51CDR			; get pseudo 6551 command register
	LSR				; shift b0 into Cb
	ROR	UPDSTAT			; shift Cb into update RSSTAT flag
	INC	BITTS			; clear RS-232 bit count
	RTS

; disable T1 interrupt

LAB_F021
	LSR	UPDSTAT			; clear update RSSTAT flag
	LDA	ENABL			; get enabled interrupts
	AND	#%10111111		; disable T1 interrupt
	STA	ENABL			; set enabled interrupts
	RTS


;***********************************************************************************;
;
; get byte from RS-232 buffer

RSNXTIN
	LDY	RIDBS			; get index to Rx buffer start
	CPY	RIDBE			; compare with index to Rx buffer end
	BEQ	LAB_F15D		; return null if buffer empty

	INC	RIDBS			; increment index to Rx buffer start
	LDA	M51CDR			; get pseudo 6551 command register
	LSR				; shift b0 into Cb
	BCC	GETB			; branch if 3 line interface

	SEC
	LDA	RIDBE			; get index to Rx buffer end
	SBC	RIDBS			; subtract index to Rx buffer start
	BEQ	GETB			; branch if Rx buffer now empty
	BIT	FLWCTL			; test Rx flow control
	BMI	GETB			; branch if Rx flow control is enabled
	CMP	#WTRMARK		; compare with low watermark
	BCS	GETB			; branch if usage is above watermark

	; Rx buffer now has plenty of space

	LDA	VIA1PB			; get VIA 1 DRB
	ORA	#%00000010		; set RTS out
	STA	VIA1PB			; set VIA 1 DRB
	LDA	#$80			; Rx flow control enabled
	STA	FLWCTL			; set Rx flow control

GETB
	LDA	(RIBUF),Y		; get byte from RS-232 Rx buffer
	SEC
	RTS

LAB_F15D
	LDA	#$00			; return null
	CLC
	RTS


;***********************************************************************************;
;
; add byte to RS-232 buffer

RSOUTSAV
	LDY	RODBE			; get index to Tx buffer end
	INY				; + 1
	CPY	RODBS			; compare with index to Tx buffer start
	BEQ	RSOUTSAV		; loop while buffer full

	DEY				; index to available buffer byte
	STA	(ROBUF),Y		; save byte to buffer
	INC	RODBE			; increment index to Tx buffer end
	CLC
	LDA	RODBS			; get index to Tx buffer start
	SBC	RODBE			; subtract Tx buffer end + 1
	BNE	CHKINT			; branch if buffer not full

	LDA	RSSTAT2			; get RS-232 additional status byte
	ORA	#%10000000		; Tx buffer full bit
	STA	RSSTAT2			; set RS-232 additional status byte
CHKINT
	BIT	VIA1IER			; test VIA 1 IER
	BVC	RSPREPOT		; branch if T1 not enabled

	RTS

RSPREPOT
	LDA	BAUDOF			; get baud rate bit time low byte
	STA	VIA1T1CL		; set VIA 1 T1C_l
	LDA	BAUDOF+1		; get baud rate bit time high byte
	STA	VIA1T1CH		; set VIA 1 T1C_h
	LDA	#%11000000		; enable T1 interrupt
	STA	VIA1IER			; set VIA 1 IER
	LDA	#$00			; clear .A
	STA	BITTS			; clear RS-232 bit count
	JMP	SETUPDF			; prepare update RSSTAT flag and exit


#ifdef TEST_HARNESS

SETLFS  = $FFBA                 ; set logical, first and second addresses
SETNAM  = $FFBD                 ; set the filename
OPEN    = $FFC0                 ; open a logical file
CLOSE   = $FFC3                 ; close a specified logical file
CHKIN   = $FFC6                 ; open a channel for input
CHKOUT  = $FFC9                 ; open a channel for output
CLRCHN  = $FFCC                 ; close input and output channels
CHRIN   = $FFCF                 ; input character from channel (via vector)
CHROUT  = $FFD2                 ; output character to channel (via vector)
GETIN   = $FFE4                 ; get a character from an input device


;***********************************************************************************;
; Test harness

RXPEROD = 120

ASTX = $02
AETX = $03
ALF = $0A
ACR = $0D
ASYN = $16



TXIND	=	$02A2
TXENABL	=	$02A3
RXENABL =	$02A4
RXTIM	=	$02A5


TESTH
	LDA	#2
	TAX
	LDY	#0
	JSR	SETLFS
	LDA	#2		; two registers
	LDX	#<REGVAL
	LDY	#>REGVAL
	JSR	SETNAM
	JSR	OPEN		; open RS-232 channel
	JMP	TINIT

FINI
	LDA	#2
	JMP	CLOSE		; close RS-232 channel

TINIT
	LDA	#ACR
	JSR	CHROUT
	LDY	#0
	STY	TXIND
	STY	TXENABL		; disable Tx
	DEY
	STY	RXENABL		; enable Rx

LOOP				; main loop
	JSR	STOP		; check STOP key
	BEQ	FINI		; exit if STOP
	BIT	TXENABL		; test Tx enabled
	BPL	SKIPTX		; branch if disabled
	BIT	RSSTAT2		; test Tx buffer full bit
	BMI	SKIPTX		; branch if buffer full
	LDX	#2
	JSR	CHKOUT
	LDA	TXIND
	CLC
	ADC	#'A'
	JSR	RSOUTSAV	; send character to RS-232 channel
	LDX	TXIND
	INX
	CPX	#26
	BCC	SKIP
	LDA	#ALF		; send LF to RS-232 channel
	JSR	RSOUTSAV
	LDX	#0		; reset counter
SKIP	STX	TXIND
	JSR	CLRCHN

SKIPTX
	BIT	RXENABL		; test Rx enabled
	BPL	CHKTIM		; branch if disabled

	LDX	#2
	JSR	CHKIN
	JSR	GETIN		; get byte from RS-232 channel
	PHA
	JSR	CLRCHN
	PLA
	BEQ	LOOP		; loop if no byte
	CMP	#' '
	BCC	CTRLCH		; branch if control character
PRTCH	JSR	CHROUT		; output character to screen
	JMP	LOOP

CHKTIM
	LDA	TIME+2		; get clock LSB
	CMP	RXTIM		; compare with expiry
	BNE	LOOP		; continue if not reached
	SEC
	ROR	RXENABL		; enable Rx
	BMI	LOOP

CTRLCH
	CMP	#ACR
	BEQ	PRTCH		; display if CR
	CMP	#ASTX
	BNE	CTRL1
	SEC
	ROR	TXENABL		; enable Tx
	BMI	LOOP
CTRL1	CMP	#AETX
	BNE	CTRL2
	LSR	TXENABL		; disable Tx
	BPL	LOOP
CTRL2	CMP	#ASYN
	BNE	LOOP
	LDA	TIME+2		; get clock LSB
	AND	#%11111100	; mask low bits
	CLC
	ADC	#RXPEROD	; add period
	STA	RXTIM
	LSR	RXENABL		; disable Rx
	JMP	LOOP

REGVAL	.BYT	$08		; 1200 baud 8N1
	.BYT	$01		; X line

#endif
